@@ -20,6 +20,11 @@ plotly_build <- function(p, registerFrames = TRUE) {
   UseMethod("plotly_build")
 }
 
+#' @export
+plotly_build.NULL <- function(...) {
+  htmltools::browsable(htmltools::div(...))
+}
+
 #' @export
 plotly_build.list <- function(p, registerFrames = TRUE) {
   plotly_build(as_widget(p))
@@ -85,6 +90,11 @@ plotly_build.plotly <- function(p, registerFrames = TRUE) {
   # Attributes should be NULL if none exist (rather than an empty list)
   if (length(p$x$attrs) == 0) p$x$attrs <- NULL
   
+  # If there is just one (unevaluated) trace, and the data is sf, add an sf layer
+  if (length(p$x$attrs) == 1 && !inherits(p$x$attrs[[1]], "plotly_eval") && is_sf(plotly_data(p))) {
+    p <- add_sf(p)
+  }
+  
   # If type was not specified in plot_ly(), it doesn't create a trace unless
   # there are no other traces
   if (is.null(p$x$attrs[[1]][["type"]]) && length(p$x$attrs) > 1) {
@@ -99,53 +109,59 @@ plotly_build.plotly <- function(p, registerFrames = TRUE) {
   if (is_mapbox(p) || is_geo(p)) {
     p <- geo2cartesian(p)
     attrsToEval <- lapply(attrsToEval, function(tr) {
-      if (!grepl("scatter|choropleth", tr[["type"]] %||% "scatter")) {
-        stop("Cant add a '", tr[["type"]], "' trace to a map object", call. = FALSE)
+      type <- tr[["type"]] %||% "scatter"
+      if (!grepl("scatter|choropleth", type)) {
+        warning("Can't add a '", type, "' trace to a map object", call. = FALSE)
+      }
+      if (is_mapbox(p)) {
+        tr[["type"]] <- if (!is.null(tr[["z"]])) "choroplethmapbox" else "scattermapbox"
       }
-      if (is_mapbox(p)) tr[["type"]] <- "scattermapbox"
       if (is_geo(p)) {
         tr[["type"]] <- if (!is.null(tr[["z"]])) "choropleth" else "scattergeo"
       }
       tr
     })
   }
-  
+
   dats <- Map(function(x, y) {
     
-    # perform the evaluation
+    # grab the data for this trace
     dat <- plotly_data(p, y)
+    
+    # formula/symbol/attribute evaluation
     trace <- structure(
       rapply(x, eval_attr, data = dat, how = "list"),
       class = oldClass(x)
     )
     
-    # attach crosstalk info, if necessary
-    if (crosstalk_key() %in% names(dat)) {
-      trace[["key"]] <- trace[["key"]] %||% dat[[crosstalk_key()]]
-      trace[["set"]] <- trace[["set"]] %||% attr(dat, "set")
-    }
-    
-    # if appropriate, tack on a group index
-    grps <- tryCatch(
-      as.character(dplyr::groups(dat)),
-      error = function(e) character(0)
-    )
-    if (length(grps) && any(lengths(trace) == NROW(dat))) {
-      trace[[".plotlyGroupIndex"]] <- interaction(dat[, grps, drop = F])
-    }
-    
     # determine trace type (if not specified, can depend on the # of data points)
     # note that this should also determine a sensible mode, if appropriate
     trace <- verify_type(trace)
     # verify orientation of boxes/bars
     trace <- verify_orientation(trace)
+    # supply sensible defaults based on trace type
+    trace <- coerce_attr_defaults(trace, p$x$layout)
+    
+    
+    
+    # attach crosstalk info, if necessary
+    if (crosstalk_key() %in% names(dat) && isTRUE(trace[["inherit"]] %||% TRUE)) {
+      trace[["key"]] <- trace[["key"]] %||% dat[[crosstalk_key()]]
+      trace[["set"]] <- trace[["set"]] %||% attr(dat, "set")
+    }
+    
+    # if appropriate, tack on a group index
+    grps <- if (has_group(trace)) tryNULL(dplyr::group_vars(dat))
+    if (length(grps) && any(lengths(trace) == NROW(dat))) {
+      trace[[".plotlyGroupIndex"]] <- interaction(dat[, grps, drop = F])
+    }
     
     # add sensible axis names to layout
     for (i in c("x", "y", "z")) {
       nm <- paste0(i, "axis")
       idx <- which(names(trace) %in% i)
       if (length(idx) == 1) {
-        title <- deparse2(x[[idx]])
+        title <- default(deparse2(x[[idx]]))
         if (is3d(trace$type) || i == "z") {
           p$x$layout$scene[[nm]]$title <<- p$x$layout$scene[[nm]]$title %||% title
         } else {
@@ -176,11 +192,13 @@ plotly_build.plotly <- function(p, registerFrames = TRUE) {
       dataArrayAttrs, special_attrs(trace), npscales(), "frame",
       # for some reason, text isn't listed as a data array in some traces
       # I'm looking at you scattergeo...
-      ".plotlyGroupIndex", "text", "key"
+      ".plotlyGroupIndex", "text", "key", "fillcolor", "name", "legendgroup"
     )
     tr <- trace[names(trace) %in% allAttrs]
     # TODO: does it make sense to "train" matrices/2D-tables (e.g. z)?
     tr <- tr[vapply(tr, function(x) is.null(dim(x)) && is.atomic(x), logical(1))]
+    # white-list customdata as this can be a non-atomic vector
+    tr$customdata <- trace$customdata
     builtData <- tibble::as_tibble(tr)
     # avoid clobbering I() (i.e., variables that shouldn't be scaled)
     for (i in seq_along(tr)) {
@@ -192,16 +210,12 @@ plotly_build.plotly <- function(p, registerFrames = TRUE) {
       isAsIs <- vapply(builtData, function(x) inherits(x, "AsIs"), logical(1))
       isDiscrete <- vapply(builtData, is.discrete, logical(1))
       # note: can only have one linetype per trace
-      isSplit <- names(builtData) %in% c("split", "linetype", "frame") |
+      isSplit <- names(builtData) %in% c("split", "linetype", "frame", "fillcolor", "name") |
         !isAsIs & isDiscrete & names(builtData) %in% c("symbol", "color")
       if (any(isSplit)) {
         paste2 <- function(x, y) if (identical(x, y)) x else paste(x, y, sep = br())
         splitVars <- builtData[isSplit]
-        traceIndex <- Reduce(paste2, splitVars)
-        if (!is.null(trace$name)) {
-          traceIndex <- paste2(traceIndex, trace$name)
-        }
-        builtData[[".plotlyTraceIndex"]] <- traceIndex
+        builtData[[".plotlyTraceIndex"]] <- Reduce(paste2, splitVars)
         # in registerFrames() we need to strip the frame from .plotlyTraceIndex
         # so keep track of which variable it is...
         trace$frameOrder <- which(names(splitVars) %in% "frame")
@@ -237,13 +251,11 @@ plotly_build.plotly <- function(p, registerFrames = TRUE) {
       )
       builtData <- train_data(builtData, trace)
       trace[[".plotlyVariableMapping"]] <- names(builtData)
-      
       # copy over to the trace data
       for (i in names(builtData)) {
         trace[[i]] <- builtData[[i]]
       }
     }
-    
     # TODO: provide a better way to clean up "high-level" attrs
     trace[c("ymin", "ymax", "yend", "xend")] <- NULL
     trace[lengths(trace) > 0]
@@ -263,7 +275,7 @@ plotly_build.plotly <- function(p, registerFrames = TRUE) {
   
   # insert NAs to differentiate groups
   traces <- lapply(traces, function(x) {
-    d <- data.frame(x[names(x) %in% x$.plotlyVariableMapping], stringsAsFactors = FALSE)
+    d <- tibble::as_tibble(x[names(x) %in% x$.plotlyVariableMapping])
     d <- group2NA(
       d, if (has_group(x)) ".plotlyGroupIndex",
       ordered = if (inherits(x, "plotly_line")) "x",
@@ -271,29 +283,37 @@ plotly_build.plotly <- function(p, registerFrames = TRUE) {
     )
     for (i in x$.plotlyVariableMapping) {
       # try to reduce the amount of data we have to send for non-positional scales
-      x[[i]] <- structure(
-        if (i %in% npscales()) uniq(d[[i]]) else d[[i]],
-        class = oldClass(x[[i]])
-      )
+      entry <- if (i %in% npscales()) uniq(d[[i]]) else d[[i]]
+      if (is.null(entry)) {
+        x[[i]] <- NULL  
+      } else {
+        x[[i]] <- structure(entry, class = oldClass(x[[i]]))  
+      }
     }
     x
   })
-  # "transforms" of (i.e., apply scaling to) special arguments
-  # IMPORTANT: scales are applied at the plot-level!!
-  colorTitle <- unlist(lapply(p$x$attrs, function(x) {
-    deparse2(x[["color"]] %||% x[["z"]])
-  }))
-  traces <- map_color(traces, title = paste(colorTitle, collapse = br()))
-  traces <- map_size(traces)
-  traces <- map_symbol(traces)
-  traces <- map_linetype(traces)
+  
+  # Map special plot_ly() arguments to plotly.js trace attributes.
+  # Note that symbol/linetype can modify the mode, so those are applied first
+  # TODO: use 'legends 2.0' to create legends for these discrete mappings
+  # https://github.com/plotly/plotly.js/issues/1668
+  if (length(traces)) {
+    traces <- map_symbol(traces)
+    traces <- map_linetype(traces)
+    traces <- map_size(traces)
+    traces <- map_size(traces, stroke = TRUE) #i.e., span
+    colorTitle <- unlist(lapply(p$x$attrs, function(x) { deparse2(x[["color"]] %||% x[["z"]]) }))
+    strokeTitle <- unlist(lapply(p$x$attrs, function(x) deparse2(x[["stroke"]])))
+    traces <- map_color(traces, title = paste(colorTitle, collapse = br()), colorway = colorway(p))
+    traces <- map_color(traces, stroke = TRUE, title = paste(strokeTitle, collapse = br()), colorway = colorway(p))
+  }
   
   for (i in seq_along(traces)) {
     # remove special mapping attributes
     mappingAttrs <- c(
-      "alpha", npscales(), paste0(npscales(), "s"),
+      "alpha", "alpha_stroke", npscales(), paste0(npscales(), "s"),
       ".plotlyGroupIndex", ".plotlyMissingIndex",
-      ".plotlyTraceIndex", ".plotlyVariableMapping"
+      ".plotlyTraceIndex", ".plotlyVariableMapping", "inherit"
     )
     for (j in mappingAttrs) {
       traces[[i]][[j]] <- NULL
@@ -316,14 +336,10 @@ plotly_build.plotly <- function(p, registerFrames = TRUE) {
   p <- supply_defaults(p)
   
   # attribute naming corrections for "geo-like" traces
-  p$x$data <- lapply(p$x$data, function(tr) {
-    if (isTRUE(tr[["type"]] %in% c("scattermapbox", "scattergeo"))) {
-      tr[["lat"]] <- tr[["lat"]] %||% tr[["y"]]
-      tr[["lon"]] <- tr[["lon"]] %||% tr[["x"]]
-      tr[c("x", "y")] <- NULL
-    }
-    tr
-  })
+  p <- cartesian2geo(p)
+  
+  # Compute sensible bounding boxes for each mapbox/geo subplot
+  p <- fit_bounds(p)
   
   # polar charts don't like null width/height keys
   if (is.null(p$x$layout[["height"]])) p$x$layout[["height"]] <- NULL
@@ -343,6 +359,12 @@ plotly_build.plotly <- function(p, registerFrames = TRUE) {
   p <- verify_hovermode(p)
   # try to convert to webgl if toWebGl was used
   p <- verify_webgl(p)
+  # throw warning if webgl is being used in shinytest
+  # currently, shinytest won't rely this warning, but it should
+  # https://github.com/rstudio/shinytest/issues/146
+  if (isTRUE(getOption("shiny.testmode"))) {
+    if (is.webgl(p)) warning("shinytest can't currently render WebGL-based graphics.")
+  }
   # crosstalk dynamically adds traces, meaning that a legend could be dynamically
   # added, which is confusing. So here we populate a sensible default.
   p <- verify_showlegend(p)
@@ -355,13 +377,29 @@ plotly_build.plotly <- function(p, registerFrames = TRUE) {
     p <- registerFrames(p, frameMapping = frameMapping)
   }
   
+  # set the default plotly.js events to register in shiny
+  p <- shiny_defaults_set(p)
+  
   p <- verify_guides(p)
   
+  # verify colorscale attributes are in a sensible data structure
+  p <- verify_colorscale(p)
+  
   # verify plot attributes are legal according to the plotly.js spec
   p <- verify_attr_names(p)
   # box up 'data_array' attributes where appropriate
   p <- verify_attr_spec(p)
   
+  # make sure we're including mathjax (if TeX() is used)
+  p <- verify_mathjax(p)
+  
+  # if a partial bundle was specified, make sure it supports the visualization
+  p <- verify_partial_bundle(p)
+  
+  # scattergl currently doesn't render in RStudio on Windows
+  # https://github.com/ropensci/plotly/issues/1214
+  p <- verify_scattergl_platform(p)
+  
   # make sure plots don't get sent out of the network (for enterprise)
   p$x$base_url <- get_domain()
   p
@@ -387,7 +425,10 @@ registerFrames <- function(p, frameMapping = NULL) {
   # remove frames from the trace names
   for (i in seq_along(p$x$data)) {
     tr <- p$x$data[[i]]
-    if (length(tr[["name"]]) != 1) next
+    if (length(tr[["name"]]) != 1) {
+      p$x$data[[i]]$frameOrder <- NULL
+      next
+    } 
     nms <- strsplit(as.character(tr[["name"]]), br())[[1]]
     idx <- setdiff(seq_along(nms), tr$frameOrder %||% 0)
     p$x$data[[i]]$name <- if (length(idx)) paste(nms[idx], collapse = br()) else NULL
@@ -443,12 +484,14 @@ registerFrames <- function(p, frameMapping = NULL) {
     d <- lapply(d, function(tr) { tr$visible <- tr$visible %||% TRUE; tr })
     
     # if this frame is missing a trace name, supply an invisible one
-    traceNamesMissing <- setdiff(frameTraceNames, sapply(d, "[[", "name"))
+    traceNamesMissing <- setdiff(frameTraceNames, unlist(lapply(d, "[[", "name")))
     for (j in traceNamesMissing) {
       idx <- vapply(p$x$data, function(tr) isTRUE(tr[["name"]] == j), logical(1))
-      idx <- which(idx)[[1]]
-      invisible <- modify_list(p$x$data[[idx]], list(visible = FALSE))
-      d <- c(d, list(invisible))
+      if (any(idx)){
+        idx <- which(idx)[[1]]
+        invisible <- modify_list(p$x$data[[idx]], list(visible = FALSE))
+        d <- c(d, list(invisible))
+      }
     }
     p$x$frames[[i]] <- list(
       name = as.character(format(nm)),
@@ -474,11 +517,9 @@ registerFrames <- function(p, frameMapping = NULL) {
   })
   
   # retrain color defaults
-  p$x$data <- retrain_color_defaults(p$x$data)
+  p$x$data <- colorway_retrain(p$x$data, colorway(p))
   p$x$frames <- lapply(p$x$frames, function(f) {
-    f$data <- retrain_color_defaults(
-      f$data, colorDefaults = traceColorDefaults()[f$traces + 1]
-    )
+    f$data <- colorway_retrain(f$data, colorway(p)[f$traces + 1])
     f
   })
   
@@ -525,32 +566,97 @@ train_data <- function(data, trace) {
     idx2 <- seq.int(2, NROW(dat), by = 2)
     dat[idx2, "x"] <- data[["xend"]]
     dat[idx2, "y"] <- data[["yend"]]
-    data <- dplyr::group_by_(dat, ".plotlyGroupIndex", add = TRUE)
+    data <- group_by_add(dat, !!rlang::sym(".plotlyGroupIndex"))
   }
+  
   # TODO: a lot more geoms!!!
   data
 }
 
 
-map_size <- function(traces) {
-  sizeList <- lapply(traces, "[[", "size")
+map_size <- function(traces, stroke = FALSE) {
+  sizeList <- lapply(traces, "[[", if (stroke) "span" else "size")
   nSizes <- lengths(sizeList)
-  # if no "top-level" color is present, return traces untouched
-  if (all(nSizes == 0)) {
-    return(traces)
-  }
+  # if no "top-level" size is present, return traces untouched
+  if (all(nSizes == 0)) return(traces)
+  
   allSize <- unlist(compact(sizeList))
   if (!is.null(allSize) && is.discrete(allSize)) {
-    stop("Size must be mapped to a numeric variable\n",
-         "symbols only make sense for discrete variables", call. = FALSE)
+    stop("`size`/`width` values must be numeric .", call. = FALSE)
   }
   sizeRange <- range(allSize, na.rm = TRUE)
   
-  types <- unlist(lapply(traces, function(tr) tr$type %||% "scatter"))
-  modes <- unlist(lapply(traces, function(tr) tr$mode %||% "lines"))
-  hasMarker <- has_marker(types, modes)
-  hasLine <- has_line(types, modes)
-  hasText <- has_text(types, modes)
+  mapSize <- switch(
+    if (stroke) "span" else "size",
+    span = function(trace, sizes) {
+      type <- trace$type %||% "scatter"
+      size_ <- uniq(sizes)
+      isSingular <- length(size_) == 1
+      attrs <- Schema$traces[[type]]$attributes
+      
+      # `span` controls marker.line.width
+      if (has_attr(type, "marker")) {
+        s <- if (isSingular) size_ else if (array_ok(attrs$marker$line$width)) sizes
+        trace$marker$line <- modify_list(list(width = default(s)), trace$marker$line)
+      }
+      
+      # `span` controls error_[x/y].thickness 
+      for (attr in c("error_y", "error_x")) {
+        if (!has_attr(type, attr)) next
+        s <- if (isSingular) size_ else if (array_ok(attrs[[attr]]$thickness)) sizes
+        trace[[attr]] <- modify_list(list(thickness = default(s)), trace[[attr]])
+      }
+      
+      # When fill exists, `span` controls line.width
+      if (has_fill(trace) && has_attr(type, "line")) {
+        s <- if (isSingular) size_ else if (array_ok(attrs$line$width)) sizes else NA
+        if (is.na(s)) {
+          warning("`line.width` does not currently support multiple values.", call. = FALSE)
+        } else {
+          trace[["line"]] <- modify_list(list(width = default(s)), trace[["line"]])
+        }
+      }
+      
+      trace
+    }, 
+    size = function(trace, sizes) {
+      type <- trace$type %||% "scatter"
+      size_ <- uniq(sizes)
+      isSingular <- length(size_) == 1
+      attrs <- Schema$traces[[type]]$attributes
+      
+      # `size` controls marker.size (note 'bar' traces have marker but not marker.size)
+      # TODO: always ensure an array? https://github.com/ropensci/plotly/pull/1176
+      if (has_attr(type, "marker") && "size" %in% names(attrs$marker)) {
+        s <- if (isSingular) size_ else if (array_ok(attrs$marker$size)) sizes
+        trace$marker <- modify_list(list(size = default(s), sizemode = default("area")), trace$marker)
+      }
+      
+      # `size` controls textfont.size
+      if (has_attr(type, "textfont")) {
+        s <- if (isSingular) size_ else if (array_ok(attrs$textfont$size)) sizes
+        trace$textfont <- modify_list(list(size = default(s)), trace$textfont)
+      }
+      
+      # `size` controls error_[x/y].width 
+      for (attr in c("error_y", "error_x")) {
+        if (!has_attr(type, attr)) next
+        s <- if (isSingular) size_ else if (array_ok(attrs[[attr]]$width)) sizes
+        trace[[attr]] <- modify_list(list(width = default(s)), trace[[attr]])
+      }
+      
+      # if fill does not exist, `size` controls line.width
+      if (!has_fill(trace) && has_attr(type, "line")) {
+        s <- if (isSingular) size_ else if (array_ok(attrs$line$width)) sizes else NA
+        if (is.na(s)) {
+          warning("`line.width` does not currently support multiple values.", call. = FALSE)
+        } else {
+          trace[["line"]] <- modify_list(list(width = default(s)), trace[["line"]])
+        }
+      }
+      trace
+    }
+  )
   
   for (i in which(nSizes > 0)) {
     s <- sizeList[[i]]
@@ -558,145 +664,216 @@ map_size <- function(traces) {
     sizeI <- if (isConstant) {
       structure(s, class = setdiff(class(s), "AsIs"))
     } else {
-      scales::rescale(s, from = sizeRange, to = traces[[1]]$sizes)
-    }
-    if (hasMarker[[i]]) {
-      traces[[i]]$marker <- modify_list(
-        list(size = sizeI, sizemode = "area"),
-        traces[[i]]$marker
-      )
-    }
-    if (hasLine[[i]]) {
-      if (!isConstant || length(sizeI) > 1) {
-        warning(
-          "plotly.js doesn't yet support line.width arrays, track this issue for progress\n",
-          "  https://github.com/plotly/plotly.js/issues/147",
-          call. = FALSE
-        )
-      }
-      traces[[i]]$line <- modify_list(
-        list(width = sizeI[1]),
-        traces[[i]]$line
-      )
-    }
-    if (hasText[[i]]) {
-      if (!isConstant || length(sizeI) > 1) {
-        warning(
-          "plotly.js doesn't yet support textfont.size arrays",
-          call. = FALSE
-        )
-      }
-      traces[[i]]$textfont <- modify_list(
-        list(size = sizeI[1]),
-        traces[[i]]$textfont
-      )
+      to <- if (stroke) traces[[1]][["spans"]] else traces[[1]][["sizes"]]
+      scales::rescale(s, from = sizeRange, to = to)
     }
+    traces[[i]] <- mapSize(traces[[i]], sizeI)
   }
   traces
 }
 
 # appends a new (empty) trace to generate (plot-wide) colorbar/colorscale
-map_color <- function(traces, title = "", na.color = "transparent") {
-  color <- lapply(traces, function(x) {
-    x[["color"]] %||% if (grepl("histogram2d", x[["type"]])) c(0, 1) else if (has_attr(x[["type"]], "colorscale")) x[["z"]] else NULL
-  })
+map_color <- function(traces, stroke = FALSE, title = "", colorway, na.color = "transparent") {
+  color <- if (stroke) {
+    lapply(traces, function(x) { x[["stroke"]] %||% x[["color"]] })
+  } else {
+    lapply(traces, function(x) { x[["color"]] %||% if (grepl("histogram2d", x[["type"]])) c(0, 1) else if (has_attr(x[["type"]], "colorscale")) x[["surfacecolor"]] %||% x[["z"]] })
+  }
+  alphas <- if (stroke) {
+    vapply(traces, function(x) x$alpha_stroke %||% 1, numeric(1)) 
+  } else {
+    vapply(traces, function(x) x[["alpha"]] %||% 1, numeric(1)) 
+  }
+  
   isConstant <- vapply(color, function(x) inherits(x, "AsIs") || is.null(x), logical(1))
   isNumeric <- vapply(color, is.numeric, logical(1)) & !isConstant
   isDiscrete <- vapply(color, is.discrete, logical(1)) & !isConstant
-  if (any(isNumeric & isDiscrete)) {
-    stop("Can't have both discrete and numeric color mappings", call. = FALSE)
-  }
+  if (any(isNumeric & isDiscrete)) stop("Can't have both discrete and numeric color mappings", call. = FALSE)
+  uniqColor <- lapply(color, uniq)
+  isSingular <- lengths(uniqColor) == 1
+  
   # color/colorscale/colorbar attribute placement depends on trace type and marker mode
-  types <- unlist(lapply(traces, function(tr) tr$type %||% "scatter"))
-  modes <- unlist(lapply(traces, function(tr) tr$mode %||% "lines"))
-  hasMarker <- has_marker(types, modes)
+  # TODO: remove these and make numeric colorscale mapping more like the rest
+  types <- vapply(traces, function(tr) tr$type %||% "scatter", character(1))
+  modes <- vapply(traces, function(tr) tr$mode %||% "lines", character(1))
   hasLine <- has_line(types, modes)
+  hasLineColor <- has_color_array(types, "line")
   hasText <- has_text(types, modes)
-  hasZ <- has_attr(types, "colorscale") &
+  hasTextColor <- has_color_array(types, "text")
+  hasZ <- has_attr(types, "colorscale") & !stroke &
     any(vapply(traces, function(tr) {
       !is.null(tr[["z"]]) || grepl("histogram2d", tr[["type"]])
     }, logical(1)))
   
-  colorDefaults <- traceColorDefaults()
-  for (i in which(isConstant)) {
-    # https://github.com/plotly/plotly.js/blob/c83735/src/plots/plots.js#L58
-    idx <- i %% length(colorDefaults)
-    if (idx == 0) idx <- 10
-    col <- color[[i]] %||% colorDefaults[[idx]]
-    alpha <- traces[[i]]$alpha %||% 1
-    rgb <- toRGB(col, alpha)
-    # we need some way to identify pre-specified defaults in subplot to retrain them
-    if (is.null(color[[i]])) attr(rgb, "defaultAlpha") <- alpha
-    obj <- if (hasLine[[i]]) "line" else if (hasMarker[[i]]) "marker" else if (hasText[[i]]) "textfont"
-    if (is.null(obj)) next
-    traces[[i]][[obj]] <- modify_list(list(color = rgb), traces[[i]][[obj]])
-    traces[[i]][[obj]] <- modify_list(list(fillcolor = rgb), traces[[i]][[obj]])
+  # IDEA - attach color codes whether they make sense, unless there is a 
+  # vector of color codes and the target is a constant
+  mapColor <- switch(
+    if (stroke) "stroke" else "fill",
+    stroke = function(trace, rgba, is_colorway = FALSE) {
+      type <- trace$type %||% "scatter"
+      rgba_ <- uniq(rgba)
+      isSingular <- length(rgba_) == 1
+      attrs <- Schema$traces[[type]]$attributes
+      default_ <- if (is_colorway) function(x) prefix_class(default(x), "colorway") else default
+      
+      if (has_attr(type, "marker")) {
+        col <- if (isSingular) rgba_ else if (array_ok(attrs$marker$line$color)) rgba
+        trace$marker$line <- modify_list(list(color = default_(col)), trace$marker$line)
+      }
+      if (has_fill(trace)) {
+        col <- if (isSingular) rgba_ else if (array_ok(attrs$line$color)) rgba
+        if (is.null(col)) {
+          warning("`line.color` does not currently support multiple values.", call. = FALSE)
+        } else {
+          trace$line <- modify_list(list(color = default_(col)), trace$line)
+        }
+      }
+      trace
+    },
+    fill = function(trace, rgba, is_colorway = FALSE) {
+      type <- trace$type %||% "scatter"
+      rgba_ <- uniq(rgba)
+      isSingular <- length(rgba_) == 1
+      attrs <- Schema$traces[[type]]$attributes
+      default_ <- if (is_colorway) function(x) prefix_class(default(x), "colorway") else default
+      
+      # `color` controls marker.color, textfont.color, error_[x/y].color
+      # TODO: any more attributes that make sense to include here?
+      for (attr in c("marker", "textfont", "error_y", "error_x")) {
+        if (!has_attr(type, attr)) next
+        if (is_colorway && "textfont" == attr) next
+        col <- if (isSingular) rgba_ else if (array_ok(attrs[[attr]]$color)) rgba else NA
+        if (is.na(col)) {
+          warning("`", attr, ".color` does not currently support multiple values.", call. = FALSE)
+        } else {
+          trace[[attr]] <- modify_list(list(color = default_(col)), trace[[attr]])
+        }
+      }
+      
+      # If trace has fill, `color` controls fillcolor; otherwise line.color
+      if (has_fill(trace)) {
+        if (!isSingular) warning("Only one fillcolor per trace allowed", call. = FALSE)
+        # alpha defaults to 0.5 when applied to fillcolor
+        if (is.null(trace[["alpha"]])) rgba_ <- toRGB(rgba_, 0.5)
+        if (isSingular) trace <- modify_list(list(fillcolor = default_(rgba_)), trace)
+      } else if (has_attr(type, "line")) {
+        # if fill does not exist, 'color' controls line.color
+        col <- if (isSingular) rgba_ else if (array_ok(attrs$line$color)) rgba else NA
+        if (is.na(col)) {
+          warning("`line.color` does not currently support multiple values.", call. = FALSE)
+        } else {
+          trace[["line"]] <- modify_list(list(color = default_(col)), trace[["line"]])
+        }
+      }
+      trace
+    }
+  )
+  
+  # i.e., interpret values as color codes
+  if (any(isConstant)) {
+    colorCodes <- Map(`%||%`, color, rep(colorway, length.out = length(traces)))
+    colorCodes <- Map(toRGB, colorCodes[isConstant], alphas[isConstant])
+    isColorway <- lengths(color[isConstant]) == 0
+    traces[isConstant] <- Map(mapColor, traces[isConstant], colorCodes, isColorway)
+  }
+  
+  # since stroke inherits from color, it should inherit the palette too
+  palette <- if (stroke) traces[[1]][["strokes"]] %||% traces[[1]][["colors"]] else traces[[1]][["colors"]]
+  
+  if (any(isDiscrete)) {
+    # unlist() does _not_ preserve order factors
+    isOrdered <- all(vapply(color[isDiscrete], is.ordered, logical(1)))
+    lvls <- getLevels(unlist(color[isDiscrete]))
+    N <- length(lvls)
+    pal <- palette %||% if (isOrdered) viridisLite::viridis(N) else RColorBrewer::brewer.pal(N, "Set2")
+    colScale <- scales::col_factor(pal, levels = names(pal) %||% lvls, na.color = na.color)
+    color_codes <- Map(function(x, y) toRGB(colScale(as.character(x)), y), color[isDiscrete], alphas[isDiscrete])
+    traces[isDiscrete] <- Map(mapColor, traces[isDiscrete], color_codes)
   }
   
   if (any(isNumeric)) {
-    palette <- traces[[1]][["colors"]] %||% viridisLite::viridis(10)
+    pal <- palette %||% viridisLite::viridis(10)
     # TODO: use ggstat::frange() when it's on CRAN?
     allColor <- unlist(color[isNumeric])
     rng <- range(allColor, na.rm = TRUE)
-    colScale <- scales::col_numeric(palette, rng, na.color = na.color)
+    colScale <- scales::col_numeric(pal, rng, na.color = na.color)
     # generate the colorscale to be shared across traces
     vals <- if (diff(rng) > 0) {
-      as.numeric(stats::quantile(allColor, probs = seq(0, 1, length.out = 25), na.rm = TRUE))
+      seq(rng[1], rng[2], length.out = 25)
     } else {
       c(0, 1)
     }
     colorScale <- matrix(
-      c(scales::rescale(vals), toRGB(colScale(vals), traces[[1]]$alpha %||% 1)),
+      c(scales::rescale(vals), toRGB(colScale(vals), alphas[[1]])),
       ncol = 2
     )
     colorObj <- list(
-      colorbar = Reduce(modify_list, lapply(traces, function(x) x$marker[["colorbar"]])) %||%
-        list(title = as.character(title), ticklen = 2),
-      cmin = rng[1],
-      cmax = rng[2],
-      colorscale = colorScale,
-      showscale = FALSE
+      colorbar = lapply(list(title = as.character(title), ticklen = 2), default),
+      cmin = default(rng[1]),
+      cmax = default(rng[2]),
+      colorscale = default(colorScale),
+      showscale = default(FALSE)
     )
     for (i in which(isNumeric)) {
+      # when colorscale is being attached to `z`, we don't need color values in
+      # colorObj, so create colorbar trace now and exit early
       if (hasZ[[i]]) {
         colorObj[c("cmin", "cmax")] <- NULL
-        colorObj[["showscale"]] <- TRUE
+        colorObj[["showscale"]] <- default(TRUE)
         traces[[i]] <- modify_list(colorObj, traces[[i]])
-        traces[[i]]$colorscale <- as_df(traces[[i]]$colorscale)
-        # sigh, contour colorscale doesn't support alpha
-        if (grepl("contour", traces[[i]][["type"]])) {
-          traces[[i]]$colorscale[, 2] <- strip_alpha(traces[[i]]$colorscale[, 2])
-        }
         traces[[i]] <- structure(traces[[i]], class = c("plotly_colorbar", "zcolor"))
         next
       }
-      colorObj$color <- color[[i]]
-      if (hasLine[[i]]) {
-        if (types[[i]] %in% c("scatter", "scattergl")) {
-          warning("Numeric color variables cannot (yet) be mapped to lines.\n",
-                  " when the trace type is 'scatter' or 'scattergl'.\n", call. = FALSE)
-          traces[[i]]$mode <- paste0(traces[[i]]$mode, "+markers")
-          hasMarker[[i]] <- TRUE
+      
+      # if trace is singular (i.e., one unique color in this trace), then there 
+      # is no need for a colorscale, and both stroke/color have relevancy
+      if (isSingular[[i]]) {
+        col <- colScale(uniq(color[[i]]))
+        traces[[i]] <- mapColor(traces[[i]], toRGB(col, alphas[[i]]))
+      } else {
+        
+        colorObj$color <- default(color[[i]])
+        
+        if (stroke) {
+          traces[[i]]$marker$line <- modify_list(colorObj, traces[[i]]$marker$line)
         } else {
-          # scatter3d supports data arrays for color
-          traces[[i]][["line"]] <- modify_list(colorObj, traces[[i]][["line"]])
-          traces[[i]]$marker$colorscale <- as_df(traces[[i]]$marker$colorscale)
+          traces[[i]]$marker <- modify_list(colorObj, traces[[i]]$marker)
         }
-      }
-      if (hasMarker[[i]]) {
-        traces[[i]][["marker"]] <- modify_list(colorObj, traces[[i]][["marker"]])
-        traces[[i]]$marker$colorscale <- as_df(traces[[i]]$marker$colorscale)
-      }
-      if (hasText[[i]]) {
-        warning("Numeric color variables cannot (yet) be mapped to text.\n",
-                "Feel free to make a feature request \n",
-                "https://github.com/plotly/plotly.js", call. = FALSE)
+        
+        if (hasLine[[i]]) {
+          if (hasLineColor[[i]]) {
+            traces[[i]]$line <- modify_list(colorObj, traces[[i]]$line)
+          } else {
+            warning("line.color doesn't (yet) support data arrays", call. = FALSE)
+          }
+        }
+        
+        if (hasText[[i]]) {
+          if (hasTextColor[[i]]) {
+            traces[[i]]$textfont <- modify_list(colorObj, traces[[i]]$textfont)
+          } else {
+            warning("textfont.color doesn't (yet) support data arrays", call. = FALSE)
+          }
+        }
+        
+        # TODO: how to make the summary stat (mean) customizable?
+        if (has_fill(traces[[i]])) {
+          warning("Only one fillcolor per trace allowed", call. = FALSE)
+          col <- toRGB(colScale(mean(colorObj$color, na.rm = TRUE)), alphas[[i]])
+          if (is.null(traces[[i]][["alpha"]])) col <- toRGB(col, 0.5)
+          traces[[i]] <- modify_list(list(fillcolor = col), traces[[i]])
+        }
+        
       }
     }
+  
+    # exit early if no additional colorbar trace is needed
     if (any(hasZ)) return(traces)
+    if (stroke && sum(lengths(lapply(traces, "[[", "stroke"))) == 0) return(traces)
+    
     # add an "empty" trace with the colorbar
     colorObj$color <- rng
-    colorObj$showscale <- TRUE
+    colorObj$showscale <- default(TRUE)
     colorBarTrace <- list(
       x = range(unlist(lapply(traces, "[[", "x")), na.rm = TRUE),
       y = range(unlist(lapply(traces, "[[", "y")), na.rm = TRUE),
@@ -707,7 +884,7 @@ map_color <- function(traces, title = "", na.color = "transparent") {
       showlegend = FALSE,
       marker = colorObj
     )
-    # yay for consistency plotly.js
+    # 3D needs a z property
     if ("scatter3d" %in% types) {
       colorBarTrace$type <- "scatter3d"
       colorBarTrace$z <- range(unlist(lapply(traces, "[[", "z")), na.rm = TRUE)
@@ -722,39 +899,6 @@ map_color <- function(traces, title = "", na.color = "transparent") {
     traces[[length(traces) + 1]] <- structure(colorBarTrace, class = "plotly_colorbar")
   }
   
-  if (any(isDiscrete)) {
-    # unlist() does _not_ preserve order factors
-    isOrdered <- all(vapply(color[isDiscrete], is.ordered, logical(1)))
-    allColor <- unlist(color[isDiscrete])
-    lvls <- getLevels(allColor)
-    N <- length(lvls)
-    pal <- traces[[1]][["colors"]] %||%
-      if (isOrdered) viridisLite::viridis(N) else RColorBrewer::brewer.pal(N, "Set2")
-    colScale <- scales::col_factor(pal, levels = names(pal) %||% lvls, na.color = na.color)
-    for (i in which(isDiscrete)) {
-      rgb <- toRGB(colScale(as.character(color[[i]])), traces[[i]]$alpha %||% 1)
-      obj <- if (hasLine[[i]]) "line" else if (hasMarker[[i]]) "marker" else if (hasText[[i]]) "textfont"
-      traces[[i]][[obj]] <- modify_list(list(color = rgb), traces[[i]][[obj]])
-      # match the plotly.js default of half transparency in the fill color
-      traces[[i]][[obj]] <- modify_list(list(fillcolor = toRGB(rgb, 0.5)), traces[[i]][[obj]])
-    }
-  }
-  
-  # marker.line.color (stroke) inherits from marker.color (color)
-  # TODO: allow users to control via a `stroke`` argument
-  # to make consistent, in "filled polygons", color -> fillcolor, stroke -> line.color
-  for (i in seq_along(color)) {
-    if (!is.null(traces[[i]]$marker$color)) {
-      traces[[i]]$marker$line$color <- traces[[i]]$marker$line$color %||% "transparent"
-      for (j in c("error_x", "error_y")) {
-        if (!is.null(traces[[i]][[j]])) {
-          traces[[i]][[j]][["color"]] <- traces[[i]][[j]][["color"]] %||%
-            traces[[i]]$marker[["color"]]
-        }
-      }
-    }
-  }
-  
   traces
 }
 
@@ -786,7 +930,7 @@ map_symbol <- function(traces) {
       )
     }
     traces[[i]][["marker"]] <- modify_list(
-      list(symbol = symbols), traces[[i]][["marker"]]
+      list(symbol = default(symbols)), traces[[i]][["marker"]]
     )
     # ensure the mode is set so that the symbol is relevant
     if (!grepl("markers", traces[[i]]$mode %||% "")) {
@@ -801,9 +945,7 @@ map_linetype <- function(traces) {
   linetypeList <- lapply(traces, "[[", "linetype")
   nLinetypes <- lengths(linetypeList)
   # if no "top-level" linetype is present, return traces untouched
-  if (all(nLinetypes == 0)) {
-    return(traces)
-  }
+  if (all(nLinetypes == 0)) return(traces)
   linetype <- unlist(compact(linetypeList))
   lvls <- getLevels(linetype)
   # get a sensible default palette
@@ -828,7 +970,7 @@ map_linetype <- function(traces) {
       )
     }
     traces[[i]][["line"]] <- modify_list(
-      list(dash = dashes), traces[[i]][["line"]]
+      list(dash = default(dashes)), traces[[i]][["line"]]
     )
     # ensure the mode is set so that the linetype is relevant
     if (!grepl("lines", traces[[i]]$mode %||% "")) {
@@ -858,7 +1000,7 @@ traceify <- function(dat, x = NULL) {
   new_dat <- list()
   for (j in seq_along(lvls)) {
     new_dat[[j]] <- lapply(dat, function(y) recurse(y, n, x %in% lvls[j]))
-    new_dat[[j]]$name <- lvls[j]
+    new_dat[[j]]$name <- new_dat[[j]]$name %||% lvls[j]
   }
   return(new_dat)
 }
@@ -876,3 +1018,35 @@ supplyUserPalette <- function(default, user) {
   }
   default
 }
+
+
+# helper functions
+array_ok <- function(attr) isTRUE(tryNULL(attr$arrayOk))
+has_fill <- function(trace) {
+  trace_type <- trace[["type"]] %||% "scatter"
+  # if trace type has fillcolor, but no fill attribute, then fill is always relevant
+  has_fillcolor <- has_attr(trace_type, "fillcolor")
+  has_fill <- has_attr(trace_type, "fill")
+  if (has_fillcolor && !has_fill) return(TRUE)
+  fill <- trace[["fill"]] %||% "none"
+  if (has_fillcolor && isTRUE(fill != "none")) return(TRUE)
+  FALSE
+}
+
+# ensure we've set a sensible trace defaults
+# based on the trace type
+coerce_attr_defaults <- function(trace, layout) {
+  # if user has specified stroke, make sure the span 
+  # defaults to something greater than 0 (so they can see the stroke!)
+  if (length(trace[["stroke"]]) && !is.default(trace[["stroke"]])) {
+    trace$span <- trace[["span"]] %||% default(I(1))
+  }
+  if (trace[["type"]] %in% c("sunburst", "pie", "treemap")) {
+    # As of v1.46.1, paper_bgcolor defaults to '#fff' which
+    # col2rgb() can't parse, but expands to '#ffffff'
+    # https://stackoverflow.com/a/2899224/1583084
+    bgcolor <- layout$paper_bgcolor %||% "#ffffff"
+    trace$stroke <- trace[["stroke"]] %||% default(I(bgcolor))
+  }
+  trace
+}
